#include <linux/autoconf.h>
#include <linux/version.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/proc_fs.h>
#include <linux/file.h>
#include <linux/mm.h>
#include <linux/pagemap.h>
#include <linux/slab.h>
#include <asm/uaccess.h>
#include <asm/unistd.h>
#include <asm/io.h>

#include "page_man.h"

extern void di_printk(char *fmt, ...);

// Initialize structures
int InitPageManager(areas_data_t* Areas)
{
        di_printk("Init page manager 0x%x", (unsigned)Areas);
        memset((void*)Areas, 0, sizeof(areas_data_t));
 return SUCCESS;
};

// Delete all mappings, free all pages and deinitialize structures
int EndPageManager(areas_data_t* Areas)
{
        int iIndex = 0;

        di_printk("Fini page manager");
        di_printk("Areas 0x%x", (unsigned) Areas);

 for (iIndex = 0; iIndex < MAX_AREAS; iIndex++)
 {
  // free pages
  if (Areas->Pages[iIndex].uiFlags & AREA_ALLOCATED)
   if (FreeArea(Areas, iIndex) != SUCCESS)
   {
    di_printk("page_man: Error unmapping area %u descriptor 0x%x\n", iIndex, Areas->Pages[iIndex].UserAreaDesc);
    break;
   }

  // Finally set to unused
  memset((void*)&Areas->Pages[0], 0, sizeof(page_repository_t));
 }
 
 if (iIndex != MAX_AREAS)
  return GENERIC_ERROR; 
 else
  return SUCCESS;
};

// Allocate area ( return area_no, descriptor used for debug and/or area managing )
int AllocateArea(areas_data_t* Areas, unsigned int size, unsigned int AreaDescriptor)
{
 int retVal = SUCCESS;
 unsigned int pages_to_alloc = (size + PAGE_SIZE - 1) / PAGE_SIZE;
 unsigned int iIndex = 0;
        unsigned int area;
        di_printk("Areas 0x%x", (unsigned) Areas);

 area = GetFreeArea(Areas);

 di_printk("AllocateArea: size=%u ", size);

 retVal = area;
 
 if (area >= 0)
 {
  di_printk("Area %u found.", area);
         Areas->Pages[area].pPages = (struct page**)vmalloc(pages_to_alloc * sizeof(struct page));
         if (Areas->Pages[area].pPages == NULL)
         {
          retVal = MEM_ALLOC_FAILED;
   di_printk("page_man: Alloc page repository failed!\n");
  }
         else
         {
          memset((void*)Areas->Pages[area].pPages, 0, pages_to_alloc * sizeof(struct page));
   Areas->Pages[area].uiFlags |= AREA_ALLOCATED;
   for (iIndex = 0; iIndex < pages_to_alloc; iIndex ++)
   {
           struct page* page;
           page = alloc_pages_node(0 /*numa_id()*/, ALLOCATION_FLAGS, 0);
           Areas->Pages[area].pPages[iIndex] = page;
           if (Areas->Pages[area].pPages[iIndex] == NULL)
           {
                          di_printk("page_man: Alloc_page failed!\n");
            retVal = PAGE_ALLOC_FAILED;
                   // Free already allocated pages
            FreeArea(Areas, area);
            break;
           }
           set_bit(PG_reserved, Areas->Pages[area].pPages[iIndex]);
           // For rollback if we unsuccessfully alloc page
    Areas->Pages[area].uiPagesCnt++;
   }
  }
 } else
  di_printk("No free areas");

 return retVal;
};

// Get pysical address of page no.. Return 0 if can't find page
unsigned int GetPhyPage(areas_data_t* Areas, unsigned int area, unsigned int page)
{
        unsigned int retVal = 0;
        
        if (Areas->Pages[area].uiFlags == AREA_FREE)
        {
                di_printk("Trying to get phys page of %u free area", area);
         return 0;
        }

 if (page > Areas->Pages[area].uiPagesCnt)
 {
                di_printk("Trying to get phys page %u of %u area containing %u pages", page, area, Areas->Pages[area].uiPagesCnt);
  return 0;
 }
 
        retVal = ((Areas->Pages[area].pPages[page] - mem_map) << PAGE_SHIFT);

 //di_printk("get phys addr (0x%x) of page %u which is %s",retVal, page,PageHighMem(Areas->Pages[area].pPages[page])?"HIGHMEM":"LOWMEM");
 return retVal;
};

// Free area ( also unmap if mapped )
int FreeArea(areas_data_t* Areas, unsigned int area)
{
        unsigned int iIndex = 0;
        di_printk("FreeArea %u", area);
        
        if (Areas->Pages[area].uiFlags == AREA_FREE)
         return INVALID_AREA;

 // Unlock pages
 if (Areas->Pages[area].uiFlags & AREA_LOCKED)
  if (UnlockArea(Areas, iIndex) != SUCCESS)
  {
   di_printk("page_man: Error unlocking area %u descriptor 0x%x\n", iIndex, Areas->Pages[iIndex].UserAreaDesc);
   return PAGE_UNLOCK_FAILED;
  }

 di_printk("vmTest: freeing pages %u", Areas->Pages[area].uiPagesCnt);
 // We remember, that uiPagesCount points to first free struct page*.
 // So, we need to decrement it.
 Areas->Pages[area].uiPagesCnt--;
 // Free used pages
 for (;Areas->Pages[area].uiPagesCnt >= 0; Areas->Pages[area].uiPagesCnt--)
 {
         clear_bit(PG_reserved, Areas->Pages[area].pPages[Areas->Pages[area].uiPagesCnt]);
         __free_page(Areas->Pages[area].pPages[Areas->Pages[area].uiPagesCnt]);
 }
 
 di_printk("vmTest: vfree pPages memory");
 
 vfree(Areas->Pages[area].pPages);
 Areas->Pages[area].pPages = NULL;

 Areas->Pages[area].uiFlags = AREA_FREE;
 
 return SUCCESS;
};

int GetFreeArea(areas_data_t* Areas)
{
  int uiArea = 0;
  int retVal = NO_MORE_AREAS;

  while (uiArea < MAX_AREAS)
  {
        if (Areas->Pages[uiArea].uiFlags == AREA_FREE)
        {
  di_printk("GetFreeArea: Found free area %u", uiArea);
         retVal = uiArea;
         break;
        }
   uiArea++;
  };

  return retVal;
};

unsigned int LockArea(areas_data_t* Areas, unsigned int area)
{
 di_printk("page_man: LockArea not realized");
        di_printk("Areas 0x%x", (unsigned) Areas);
 return 0;
};

unsigned int UnlockArea(areas_data_t* Areas, unsigned int area)
{
 di_printk("page_man: UnlockArea not realized");
        di_printk("Areas 0x%x", (unsigned) Areas);
 return 0;
};

